/*
 * LICENSED UNDER THE LGPL 2.1,
 * http://www.gnu.org/licenses/lgpl.html
 * Also, the author promises to never sue IBM for any use of this file.
 *
 */

package net.jakubholy.testing.dbunit.embeddeddb;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.sql.SQLException;
import java.util.Properties;

import javax.sql.DataSource;

import junit.framework.Assert;
import net.jakubholy.testing.dbunit.embeddeddb.assertion.RowComparator;
import net.jakubholy.testing.dbunit.embeddeddb.impl.EnhancedDatabaseTesterDecorator;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.dbunit.DatabaseUnitException;
import org.dbunit.DatabaseUnitRuntimeException;
import org.dbunit.IDatabaseTester;
import org.dbunit.PropertiesBasedJdbcDatabaseTester;
import org.dbunit.database.DatabaseConfig;
import org.dbunit.database.IDatabaseConnection;
import org.dbunit.dataset.CompositeDataSet;
import org.dbunit.dataset.DataSetException;
import org.dbunit.dataset.IDataSet;
import org.dbunit.dataset.NoSuchTableException;
import org.dbunit.dataset.xml.XmlDataSet;
import org.dbunit.operation.DatabaseOperation;

/**
 * The core class for unit tests that need to access a database, by default
 * an embedded (Derby) database, as simply as possible.
 * It relies on DbUnit to do most of the work, including reloading the
 * database with the test date before each test.
 * <p>
 * You will want to use this class if you prefer not to subclass the
 * {@link AbstractEmbeddedDbTestCase}, for instance if using Junit 4.
 * <p>
 * The class uses an embedded Derby DB for test data.
 * The necessary tables must be already created
 * - you may achieve that by creating a DDL file {@link DatabaseCreator#DDL_FILE_PATH}
 * (normally testData/create_db_content.ddl) and either executing
 * {@link DatabaseCreator#main(String[])} or calling {@link DatabaseCreator#createAndInitializeTestDb()}.
 * The database is expected to be in ./testData/testDB/ (a path relative to the
 * folder where your unit tests are executed from), which will be taken care of
 * if you use the DatabaseCreator to create the DB from a DDL.
 * </p><p>
 * If you want to load data from another file than the default one, you can
 * either use {@link #setDataSet(IDataSet)} or override {@link #createDefaultDataSet()}.
 * </p><p>
 * The added value of this class is that it's pre-configured to interact with
 * an embedded Derby database and the only thing you need to do is to provide
 * a DDL file, execute the {@link DatabaseCreator} to create the database,
 * and enter data into the predefined data set XML file. In addition to that it
 * also provides some nice utility methods and classes such as
 * {@link RowComparator} and {@link #getEnhancedTester()} and it overrides
 * the default DbUnit setting to
 * use fully qualified table names (in the form schema.table).
 * </p>
 *
 * <h3>How to use</h3>
 * Follow the 6 easy steps at the <a href="http://sourceforge.net/apps/mediawiki/jeeutils/index.php?title=DbUnit_Test_Skeleton#How_to_use_it">
 * homepage</a>.
 * Summary:
 * <ol>
 * 	<li>Add the necessary dependencies to your project (see the web page or the embedded pom.xml)
 * 	<li>Create the folder testData/ under your project's root folder.
 * 	<li>(optional) Prepare .ddl: Copy DbUnitTestSkeleton/testData/create_db_content.ddl into the new testData/ and modify it to fit your data schema needs.
 * 	<li>(optional) Create&amp;Initialize DB: Run DatabaseCreator#main from your project's root folder to create and initialize the test DB from the .ddl file.
 * 	<li>Write a TestCase either using {@link EmbeddedDbTester} or subclassing the
 * {@link AbstractEmbeddedDbTestCase} and implement your tests there.
 * (See SimpleEmbeddedDbTest.java for inspiration.) You may want to override its getDataSet()
 * if you want to use st. else than testData/dbunit-test_data_set.xml.
 * 	<li>Prepare test data: Copy DbUnitTestSkeleton/testData/dbunit-test_data_set.xml
 * into your new testData/ and modify it to fit your data needs. Any table mentioned
 * in the file will be emptied prior to running test, if it has any rows
 * defined then those will be also inserted.
 * 	<li>Run the test.
 * </ol>
 *
 * <h3>Example of usage</h3>
 * See the <a href="http://jeeutils.svn.sourceforge.net/viewvc/jeeutils/trunk/DbUnitTestSkeleton/src/main/test/net/jakubholy/testing/dbunit/embeddeddb/SimpleEmbeddedDbTest.java?view=markup">
 * SimpleEmbeddedDbTest.java</a> which is a part of this project to see
 * a complete example.
 *
 * <h4>Verify data in DB</h4>
 * <pre><code>
 * // Inside a test class ...
 *
 * private EmbeddedDbTester testDb = new EmbeddedDbTester();
 *
 * public void setUp() {
 * 	testDb.onSetup();
 * }
 *
 * public void tearDown() {
 * 	testDb.onTearDown();
 * }
 *
 * public void testMyMethodXY() {
 * 	// ...
 * 	// after invoking the method under test:
 * 	ITable rowsToInsertTbl = testDb.getConnection().createQueryTable("rowToInsert", "SELECT MY_EXAMPLE_COLUMN FROM ... WHERE ...");
 * 	assertEquals("There shall be exactly 1 row for the given id.", 1, rowsToInsertTbl.getRowCount());
 * 	String dbVal = (String) rowsToInsertTbl.getValue(0, "MY_EXAMPLE_COLUMN");
 *  assertEquals("Value expected in the MY_EXAMPLE_COLUMN", "hello", dbVal);
 * 	}
 * </code></pre>
 *
 * <h3>Custom configuration and customizations</h3>
 * By default we use an embedded Derby database stored at a specific location
 * as defined by this class' constants. You may change the driver, database etc.
 * by providing a configuration of your own in a properties file, see
 * {@link #CUSTOM_CONFIG_FILE}.
 * <p>
 * You may also change what data set is used for initializing the database by
 * either overriding {@link #createDefaultDataSet()} or setting the
 * desired data set with {@link #setDataSet(IDataSet)}.
 *
 * See <a href="http://sourceforge.net/apps/mediawiki/jeeutils/index.php?title=DbUnit_Test_Skeleton">DbUnit_Test_Skeleton subproject of jeeutils at SourceForge</a>
 *
 * @see #onSetup()
 * @see #getEnhancedTester()
 * @see #createCheckerForSelect(String)
 * @see #getConnection()
 * @see #getDataSource()
 *
 * @see DatabaseCreator#createAndInitializeTestDb()
 * @see DatabaseCreator#createDbSchemaFromDdl(java.sql.Connection)
 * @see DatabaseCreator#DDL_FILE_PATH
 * @see DatabaseCreator#main(String[])
 *
 * @author jholy
 * @since 1.1.0
 */
public class EmbeddedDbTester implements IDatabaseTester {

	/**
	 * File's Subversion info (version etc.).
	 * It's replacement upon commit must be enabled in svn properties.
	 */
	public static final String SVN_ID = "$Id: EmbeddedDbTester.java 68 2010-02-15 10:36:31Z malyvelky $";

	private static final Log LOG = LogFactory.getLog(EmbeddedDbTester.class);

	//############# TEST DATA CONFIG ################
    /** Name of the folder storing our test data set XML, DB initialization DDL and the Derby db itself. */
	public static final String TEST_DATA_FOLDER = "testData";

	/**
	 * Name of our test data set XML file loaded by default into the database
	 * by this class unless overridden by a subclass.
	 * The file should be stored in the folder {@link #TEST_DATA_FOLDER}.
	 */
    public static final String DBUNIT_TEST_DATA_SET_NAME = "dbunit-test_data_set.xml"; // NOPMD
    //############# /TEST DATA CONFIG ################

	//############# JDBC CONNECTION INFO ################
	/**
	 * Path to the directory storing the test derby DB.
	 * May be absolute or relative to the test execution directory.
	 */
    private static final String DERBY_DB_PATH = TEST_DATA_FOLDER + File.separator + "testDB";

	/** Test DB connection info - url. */
	private static final String DEFAULT_JDBC_URL = "jdbc:derby:" + DERBY_DB_PATH;

	/** Test DB connection info - driver class name. */
	private static final String DEFAULT_JDBC_DRIVER = "org.apache.derby.jdbc.EmbeddedDriver";

	/** Test DB connection info - user name. */
	private static final String DEFAULT_JDBC_USERNAME = "sa";

	/** Test DB connection info - password. */
	private static final String DEFAULT_JDBC_PASSWORD = "";
	//############# /JDBC CONNECTION INFO ################

    /** DBUnit internal stuff that takes care of connecting to the DB. */
    private IEnhancedDatabaseTester tester;	// NOPMD

    private transient IDataSet testDataSet = null;

    /**
     * An optional file for defining other then the default connection to a test
     * database, it may include any of the following properties:
     * <ul>
     * 	<li> {@link PropertiesBasedJdbcDatabaseTester#DBUNIT_CONNECTION_URL}
     * 	<li> {@link PropertiesBasedJdbcDatabaseTester#DBUNIT_DRIVER_CLASS}
     * 	<li> {@link PropertiesBasedJdbcDatabaseTester#DBUNIT_PASSWORD}
     * 	<li> {@link PropertiesBasedJdbcDatabaseTester#DBUNIT_USERNAME}
     * </ul>
     * If a property isn't defined in the file then the value of the
     * appropriate DEFAULT_JDBC_* constant is taken.
     */
    public static final String CUSTOM_CONFIG_FILE = "dbunit-embedded.properties";

    private final Properties connectionProps;	// NOPMD

	private transient DatabaseOperation setUpOperation = DatabaseOperation.CLEAN_INSERT;

	private transient DatabaseOperation tearDownOperation = DatabaseOperation.NONE;

    /**
     * Create a new embedded DB tester ready to use, optionally configured by
     * properties defined in the file {@value #CUSTOM_CONFIG_FILE} if
     * it's somewhere on the classpath.
     * <p>
     * Do not forget to call its {@link #onSetup()} before using it in a test
     * to modify or read the test database.
     */
    public EmbeddedDbTester() {
    	connectionProps = loadConnectionConfig();
    }

	/**
	 * Load the connection configuration from {@value #CUSTOM_CONFIG_FILE}
	 * if available. Fail-safe.
	 * @return the loaded properties (empty if not found or an exception occurred)
	 */
	private static Properties loadConnectionConfig() {
		final Properties properties = new Properties();
		final URL configUrl = EmbeddedDbTester.class.getResource(
    			"/" + CUSTOM_CONFIG_FILE);
    	if (configUrl != null) {
    		LOG.info("Loading test DB configuration from " + configUrl);
    		try {
				properties.load( configUrl.openStream() );
			} catch (IOException e) {
				LOG.warn("Failed to read DB configuration from " +
						configUrl + ": " + e, e);
			}
    	}
    	return properties;
	}


    /**
     * Create the internally used database tester, which supports fully
     * qualified table names and uses an embedded Derby database.
     * <p>
     * We also set the system properties that define the test DB connection
     * based on the getCustomJdbc* methods or the DEFAULT_JDBC_*
     * constants like {@link #DEFAULT_JDBC_URL} if the methods return null.
     */
	protected final IEnhancedDatabaseTester createAndInitDatabaseTester() {

        System.setProperty( PropertiesBasedJdbcDatabaseTester.DBUNIT_DRIVER_CLASS,
        		connectionProps.getProperty(
        				PropertiesBasedJdbcDatabaseTester.DBUNIT_DRIVER_CLASS
        				, DEFAULT_JDBC_DRIVER));
        System.setProperty( PropertiesBasedJdbcDatabaseTester.DBUNIT_CONNECTION_URL,
        		connectionProps.getProperty(
        				PropertiesBasedJdbcDatabaseTester.DBUNIT_CONNECTION_URL
        				, DEFAULT_JDBC_URL));
        System.setProperty( PropertiesBasedJdbcDatabaseTester.DBUNIT_USERNAME,
        		connectionProps.getProperty(
        				PropertiesBasedJdbcDatabaseTester.DBUNIT_USERNAME
        				, DEFAULT_JDBC_USERNAME));
        System.setProperty( PropertiesBasedJdbcDatabaseTester.DBUNIT_PASSWORD,
        		connectionProps.getProperty(
        				PropertiesBasedJdbcDatabaseTester.DBUNIT_PASSWORD
        				, DEFAULT_JDBC_PASSWORD));
        //System.setProperty( PropertiesBasedJdbcDatabaseTester.DBUNIT_SCHEMA, "MY_DEFAULT_SCHEMA" );

		final IDatabaseTester actualTester = new QualifiedNamesPropertiesTester();

		return new EnhancedDatabaseTesterDecorator(actualTester);
	} /* createDatabaseTester */

	/**
	 * A DatabaseTester that expects table names to be fully qualified, i.e.
	 * including a schema name. This makes it possible to use tables from
	 * different schemas in the same test.
	 */
	private static final class QualifiedNamesPropertiesTester extends
			PropertiesBasedJdbcDatabaseTester {
		public IDatabaseConnection getConnection() throws Exception { // NOPMD
			try {
		        final IDatabaseConnection conn = super.getConnection();
		        conn.getConfig().setFeature(DatabaseConfig.FEATURE_QUALIFIED_TABLE_NAMES, true);
		        return conn;
			} catch (SQLException e) {
				final String missingDbMsg = "Database '" + DERBY_DB_PATH +
					"' not found.";
				if (missingDbMsg.equals(e.getMessage())) {
					throw new RuntimeException("Failed to connect to the " +
							"test database, it seems it hasn't been " +
							"created yet - you can do that by executing " +
							DatabaseCreator.class + ".main(..)."
							, e);
				}
				throw e;
			}
		}
	}

	/**
	 * Get the internally used tester implementation instance, if it is not
	 * set yet it will be created first.
	 * @see #createAndInitDatabaseTester
	 */
	protected final IEnhancedDatabaseTester getWrappedTester() {
		if (tester == null) {
			tester = createAndInitDatabaseTester();
		}
		return tester;
	}

	/**
	 * With this method a subclass may replace the internally used wrapper
	 * by calling it e.g. in the constructor.
	 */
	protected final void setWrappedTester(final IEnhancedDatabaseTester actualTester) {
		this.tester = actualTester;
	}

	/**
	 * Returns an enhanced manager of the test database that provides
	 * additional utility methods, for example for replacing its data.
	 *
	 * @see IEnhancedDatabaseTester#clearTable(String)
	 * @see IEnhancedDatabaseTester#replaceDatabase(String)
	 */
	public IEnhancedDatabaseTester getEnhancedTester() {
		return getWrappedTester();
	}

	/**
     * Create the default data set used to fill the test DB by reading the
     * default XML {@value #DBUNIT_TEST_DATA_SET_NAME}. Used if no
     * custom data set is provided via {@link #setDataSet(IDataSet)}.
     * Override if you want to load data from a different file(s)
     * or if you prefer a different format of the data sat (e.g. FlatXmlDataSet).
     */
    protected IDataSet createDefaultDataSet() throws DatabaseUnitRuntimeException, DataSetException {
		return createDataSetFromFile(DBUNIT_TEST_DATA_SET_NAME);
	}

    /**
     * Create a data set from the given XML file stored in the default location
     * of {@value #TEST_DATA_FOLDER}.
     * @param xmlFileName (required) a XML file defining DbUnit data set in the testData folder
     * @return a new data set created from that file
     * @throws DatabaseUnitRuntimeException If the file cannot be found
     * @throws DataSetException If there is a problem with the data set format
     */
    public final IDataSet createDataSetFromFile(final String xmlFileName)
    throws DatabaseUnitRuntimeException, DataSetException {

    	if (xmlFileName == null) {
    		throw new IllegalArgumentException("String xmlFileName may not be null");
    	}

    	InputStream dataSetStream;

    	final String defaultPath =
    		TEST_DATA_FOLDER + File.separator + xmlFileName;
    	final File defaultFile = new File(defaultPath);

    	if (defaultFile.canRead()) {
    		try {
				dataSetStream = new FileInputStream( defaultFile );
			} catch (FileNotFoundException e) {
				// this should not happen for we tested defaultFile.canRead()
				throw new DatabaseUnitRuntimeException(e);
			}
    	} else {
    		dataSetStream = findOnClasspath(xmlFileName, defaultPath);
    	}

    	return new CompositeDataSet(new IDataSet[]{
    			new XmlDataSet(dataSetStream)
    	});
    }

	/**
	 * Find a data set file on the class-path or fail.
	 * @param xmlFileName (required) the file to search for
	 * @param defaultPath (required) the location where the file was looked for
	 * 	originally (for logging)
	 * @return
	 * @throws DatabaseUnitRuntimeException If the file cannot be found or opened
	 */
	private InputStream findOnClasspath(final String xmlFileName, final String defaultPath)
			throws DatabaseUnitRuntimeException {

		LOG.info("createDataSetFromFile(" + xmlFileName + "): Can't read " +
				"the file at the default location '" + defaultPath +
				"', trying to find it on the classpath.");

		final URL dataSetOnCpUrl = getClass().getResource(xmlFileName);

		if (dataSetOnCpUrl == null) {
			final String msg = "The data set file '" +
				xmlFileName + "' can't be  found neither in the default " +
				defaultPath + " nor on the classpath (notice that the " +
				"default search location is relative and thus depends on " +
				"the folder where you execute the tests from)";
			LOG.warn("createDataSetFromFile: " + msg);
			throw new DatabaseUnitRuntimeException(msg);
		} else {
			LOG.info("createDataSetFromFile(" + xmlFileName + "): Not " +
					" in the " + TEST_DATA_FOLDER + " but found on the " +
					"classpath at " + dataSetOnCpUrl);
			try {
				return dataSetOnCpUrl.openStream();
			} catch (IOException e) {
				throw new DatabaseUnitRuntimeException(
						"createDataSetFromFile: Failed to read the file " +
						dataSetOnCpUrl, e);
			}
		}
	}

	/**
	 * Returns a data source for accessing the underlying test database.
	 * A shortcut for {@link IEnhancedDatabaseTester#getDataSource()}.
	 * @throws DatabaseUnitRuntimeException
	 */
	public DataSource getDataSource() throws DatabaseUnitRuntimeException {
		return getWrappedTester().getDataSource();
	}

	/**
	 * Create a new RowComparator for comparing results of a SQL select over
	 * the test database with your expectations. See an example of usage in
	 * the class' JavaDoc.
	 *
	 * @param sqlSelect (required) a SQL SELECT statement on the test DB
	 * @return a new comparator loaded with results of the sqlSelect
	 * @throws DatabaseUnitRuntimeException
	 * @throws SQLException
	 * @throws Exception
	 *
	 * @see RowComparator#RowComparator(IDatabaseTester, String)
	 * @see RowComparator#assertRowCount(int)
	 * @see RowComparator#assertNext(Object[])
	 */
	public RowComparator createCheckerForSelect(final String sqlSelect) throws DatabaseUnitRuntimeException, SQLException {
		return new RowComparator(getWrappedTester(), sqlSelect);
	}

	// ####################################################### INTERFACE METHODS

	/* (non-Javadoc)
	 * @see org.dbunit.IDatabaseTester#closeConnection(org.dbunit.database.IDatabaseConnection)
	 */
	public void closeConnection(final IDatabaseConnection connection)
			throws Exception {	// NOPMD
		getWrappedTester().closeConnection(connection);
	}

	/**
	 * Returns a connection to the underlying test database, which can
	 * be also used to obtain a normal SQL connection.
	 * @see IDatabaseConnection#getConnection()
	 * @see org.dbunit.IDatabaseTester#getConnection()
	 */
	public IDatabaseConnection getConnection() throws Exception {	// NOPMD
		return getWrappedTester().getConnection();
	}

	/**
     * Initiate the database by clearing tables and filling them with data from the data set.
     * Only tables that are mentioned in the data set are cleared of data.
     * <p>
     * The default set up operation and test data sets are used unless you
     * they have been changed via their setters (only rarely needed)
     * <p>
     * We've added some better error reporting in the case that the test DB
     * hasn't been created yet.
     *
     * @see org.dbunit.DatabaseTestCase#setUp()
     */
    public void onSetup() throws Exception {	// NOPMD
        try {
            final IDatabaseTester databaseTester = getWrappedTester();
            Assert.assertNotNull( "DatabaseTester is not set", databaseTester );
            databaseTester.setSetUpOperation( setUpOperation );
            databaseTester.setDataSet( getDataSet() );
            databaseTester.onSetup();
        } catch (NoSuchTableException e) {
            throw new DatabaseUnitException("No such table exception - " +
            		"have you created & initialized the embedded DB? " +
            		"The missing table is '" + e.getMessage() + "'", e);
        } catch (Exception e) {
        	if (e instanceof SQLException) {	// NOPMD
        		LOG.warn("If the initial DB cleanup failed because of nonexistant schema " +
        			"(usually same as logged-in user name, for derby SA), set the System property " +
        			"PropertiesBasedJdbcDatabaseTester.DBUNIT_SCHEMA to an existing schema. Error: " +
        			e.getMessage(), e);
        	}
        	throw new DatabaseUnitException(e);
        }
	}

	/* (non-Javadoc)
	 * @see org.dbunit.IDatabaseTester#onTearDown()
	 */
	public void onTearDown() throws Exception {	// NOPMD
		getWrappedTester().setTearDownOperation(tearDownOperation);
		getWrappedTester().onTearDown();
	}

	/**
	 * Get the test data set used for (re-)initializing the test DB.
	 * If none has been set via {@link #setDataSet(IDataSet)} then a default
	 * D.S. is created.
	 * @return the testDataSet
	 *
	 * @see IDatabaseTester#getDataSet()
	 * @throws DatabaseUnitRuntimeException Failure creating the data set
	 */
	public IDataSet getDataSet() throws DatabaseUnitRuntimeException {
		if (testDataSet == null) {
			try {
				testDataSet = createDefaultDataSet();
			} catch (DataSetException e) {
				throw new DatabaseUnitRuntimeException(
						"Failed to create the default test data set", e);
			}
		}
		return testDataSet;
	}

	/**
	 * If you do not want the default test data set to be used
	 * (read from {@value #TEST_DATA_FOLDER}/{@value #DBUNIT_TEST_DATA_SET_NAME})
	 * but prefer some other one, you can set it via this method.
	 *
	 * @param testDataSet the data to insert into the DB before each test
	 *
	 * @see IDatabaseTester#setDataSet(IDataSet)
	 */
	public void setDataSet(final IDataSet testDataSet) {
		this.testDataSet = testDataSet;
	}

	/**
	 * This method is meaningless for this particular tester and thus throws
	 * an exception.
	 * @throws UnsupportedOperationException This tester requires table names
	 * to be always fully qualified
	 * (i.e. schema.table) and thus setting a schema has no effect
	 * and shouldn't be used.
	 * @see org.dbunit.IDatabaseTester#setSchema(java.lang.String)
	 */
	public void setSchema(final String schema) throws UnsupportedOperationException {
		throw new UnsupportedOperationException("This tester requires table" +
				" names to be always fully qualified (i.e. schema.table)" +
				"and thus setting a schema has no effect and shouldn't be used");

	}

	/**
	 * Normally you shouldn't need this.
	 * {@inheritDoc}
	 * @see org.dbunit.IDatabaseTester#setSetUpOperation(org.dbunit.operation.DatabaseOperation)
	 */
	public void setSetUpOperation(final DatabaseOperation setUpOperation) {
		this.setUpOperation = setUpOperation;
	}

	/**
	 * Normally you shouldn't need this.
	 * {@inheritDoc}
	 * @see org.dbunit.IDatabaseTester#setTearDownOperation(org.dbunit.operation.DatabaseOperation)
	 */
	public void setTearDownOperation(final DatabaseOperation tearDownOperation) {
		this.tearDownOperation = tearDownOperation;
	}

}
