/*
 * LICENSED UNDER THE LGPL 2.1,
 * http://www.gnu.org/licenses/lgpl.html
 * Also, the author promises to never sue IBM for any use of this file.
 *
 */

package net.jakubholy.testing.dbunit.embeddeddb.assertion;

import java.sql.SQLException;

import junit.framework.Assert;
import junit.framework.AssertionFailedError;

import org.dbunit.Assertion;
import org.dbunit.DBTestCase;
import org.dbunit.DatabaseTestCase;
import org.dbunit.DatabaseUnitRuntimeException;
import org.dbunit.IDatabaseTester;
import org.dbunit.dataset.Column;
import org.dbunit.dataset.DataSetException;
import org.dbunit.dataset.DefaultTable;
import org.dbunit.dataset.ITable;
import org.dbunit.dataset.ITableMetaData;

/**
 * A helper class for comparing actual and expected values of the test
 * database' rows using DbUnit. It loads data from the test database into
 * an internal result table ({@link ITable}) and provides convenient
 * methods for checking its content.
 * <p>
 * You may also consider using directly the assert methods for ITable
 * and IDataSet in {@link Assertion}, perhaps together with
 * {@link DefaultTable#DefaultTable(String, Column[])}.
 *
 * <h4>Standard usage</h4>
 * You would use it as follows from a subclass of {@link DatabaseTestCase}:
 * <pre><code>
 * final Product[] expected = new Product[] { ... };
 * RowComparator cmp = new RowComparator(getDatabaseTester()
 * 		, "select productId, price from eshop.products order by productId");
 * cmp.assertRowCount(3);
 * cmp.assertNext(new Object[]{1, 1000.00});	// requires Java 5's autoboxing
 * cmp.assertNext(new Object[]{2, 333.50});
 * cmp.assertNext("3", "99.0"); // using the convenience String comparison
 * </code></pre>
 *
 *
 * <h4>Subclassing for convenience</h4>
 * You can create a subclass to make verification of
 * your data/objects more convenient, something like
 * <pre><code>
 * private class ProductRowComparator extends RowComparator {
 *
 * 	private final Product[] expected;
 *
 * 	public ProductRowComparator(final IDatabaseTester databaseTester, final Product[] expected) {
 * 		super(databaseTester, "select productId, price from eshop.products order by productId");
 * 		this.expected = expected.clone();
 * 		java.util.Arrays.sort(this.expected); // sort by productId
 * 	}
 *
 * 	public void assertRowsEqual()  throws AssertionFailedError, DataSetException {
 * 		super.assertRowCount(expected.length);
 * 		for (int i = 0; i < expected.length; i++) {
			assertNext(new Object[]{ expected[i].getProductId(), expected[i].getPrice() });
		}
 * 	}
 *
 * }
 * </code></pre>
 * You would than use it as follows:
 * <pre><code>
 * final Product[] expected = new Product[] { ... };
 * new ProductRowComparator(getDatabaseTester(), expected).assertRowsEqual();
 * </code></pre>
 *
 */
public class RowComparator {

//	private static final Log LOG = LogFactory.getLog(RowComparator.class);

	private String[] columnNames;	// NOPMD
	private ITable resultTable;		// NOPMD
	private int currentRow = -1;	// NOPMD

	/**
	 * A new comparator that creates the internal resultTable ({@link ITable})
	 * from the provided SQL and accesses the test database via the
	 * provided tester.
	 * @param databaseTester (required) necessary to read the test DB;
	 * 	see {@link DBTestCase#getDatabaseTester()}
	 * @param sql (required) the SQL SELECT to execute
	 * @throws SQLException
	 * @throws DatabaseUnitRuntimeException
	 */
	public RowComparator(final IDatabaseTester databaseTester, final String sql) throws SQLException, DatabaseUnitRuntimeException {
		initResultTableAndColumns(databaseTester, sql);
	} /* constructor(String) */

	/**
	 * Check expected values against the provided {@link ITable}.
	 * @param resultTable (required)
	 * @throws DataSetException
	 */
	public RowComparator(final ITable resultTable) throws DataSetException {
		initResultTableAndColumns(resultTable);
	} /* constructor(ITable) */

	/**
	 * You must call {@link #initResultTableAndColumns(ITable)} or
	 * {@link #initResultTableAndColumns(IDatabaseTester, String)} if using this constructor.
	 */
	protected RowComparator() { /* for subclasses to override */ }

	/**
	 * Use the SQL to create an ITable and call {@link #initResultTableAndColumns(ITable)}.
	 * @param databaseTester
	 * @param sql (required)
	 * @throws SQLException
	 * @throws DatabaseUnitRuntimeException by {@link IDatabaseTester#getConnection()}
	 */
	protected final void initResultTableAndColumns(final IDatabaseTester databaseTester
			, final String sql) throws SQLException, DatabaseUnitRuntimeException {
		if (databaseTester == null) {
			throw new IllegalArgumentException("The argument databaseTester: IDatabaseTester may not be null.");
		}

		try {
			final ITable resultTable = databaseTester.getConnection().
				createQueryTable("rowComparatorTbl", sql);
			initResultTableAndColumns(resultTable);
		} catch (Exception e) {
			throw new DatabaseUnitRuntimeException(e);
		}

	} /* initResultTableAndColumns(String) */

	/**
	 * Set {@link #resultTable} and use its metadata to initialize {@link #columnNames}.
	 * @param resultTable (required)
	 * @throws DataSetException
	 */
	protected final void initResultTableAndColumns(final ITable resultTable) throws DataSetException {
		if (resultTable == null) {
			throw new IllegalArgumentException("ITable resultTable may not be null");
		}
		this.resultTable = resultTable;
		this.columnNames = extractColumnNames();
	} /* initResultTableAndColumns(ITable) */

	/**
	 * Extract column names from {@link #resultTable}.
	 * @throws DataSetException
	 * @see ITable#getTableMetaData()
	 * @see ITableMetaData#getColumns()
	 */
	private String[] extractColumnNames() throws DataSetException {
		final Column[] columns = resultTable.getTableMetaData().getColumns();
		final String[] columnNames = new String[columns.length];	// NOPMD

		for (int i = 0; i < columns.length; i++) {
			columnNames[i] = columns[i].getColumnName();	// NOPMD
		}
		return columnNames;
	} /* extractColumnNames */

	/**
	 * See {@link #assertNext(String, Object[])}.
	 */
	public void assertNext(final Object[] expectedValues) throws AssertionFailedError, DataSetException {
		assertNext(null, expectedValues);
	}

	/**
	 * Compare the next row (or the 1st one if first use) in this.{@link #resultTable} with
	 * the provided expected values and throw an assertion failure if they differ.
	 * <p>
	 * If the row differs in a column then an AssertionFailedError is thrown
	 * including the row number, name of the column and the expected and actual
	 * values.
	 * @param msg (optional) A message to add to the AssertionFailedErrors produced
	 * @param expectedValues (required) the values to be expected in the columns;
	 * the values must be in the same order as columns in the internal {@link ITable}.
	 *
	 * @see ITable#getTableMetaData()
	 * @see ITableMetaData#getColumns()
	 */
	public void assertNext(final String msg, final Object[] expectedValues) throws AssertionFailedError, DataSetException {

		final String errorMsg = (msg == null)? "" : msg;		// NOPMD

		++currentRow;
		Assert.assertTrue("There is no next row, the row count is " + resultTable.getRowCount()
				, currentRow < resultTable.getRowCount());

		if (expectedValues == null) {
			throw new IllegalArgumentException("Object[] expectedValues may not be null");
		}
		if (columnNames.length != expectedValues.length) {
			throw new IllegalArgumentException("columnNames.length (" + columnNames.length +
					") shall be same as expectedValues.length (" + expectedValues.length + ")");
		}

		final boolean stringComparison = expectedValues.getClass() // NOPMD
			.getComponentType().equals(String.class);

		// Do the test
		for (int i = 0; i < columnNames.length; i++) {
			Object actual = resultTable.getValue(currentRow, columnNames[i]);
			if (stringComparison && actual != null) {
				actual = actual.toString();
			}
			Assert.assertEquals(errorMsg + " (row (starting from 0) " + currentRow +
					", column '" + columnNames[i] + "')"
					, expectedValues[i], actual);
		}

	}

	/**
	 * Assert that the values of the next row are same as the three ones provided
	 * here after being converted to strings via toString().
	 * This is a convenience method that makes asserting particular values
	 * easier (because you don't need to create an Object array and wrap your
	 * primitives) if string comparison is sufficient for you.
	 * <p>
	 * To be used only if the underlying table has exactly three columns.
	 * @param expectedFirst (optional) the expected value of the first column
	 * @param expectedSecond (optional) the expected value of the 2nd column
	 * @param expectedThird (optional) the expected value of the 3rd column
	 * @throws AssertionFailedError
	 * @throws DataSetException
	 */
	public void assertNext(final String expectedFirst, final String expectedSecond, final String expectedThird) throws AssertionFailedError, DataSetException {
		assertNext(null, new String[]{expectedFirst, expectedSecond, expectedThird});
	}

	/**
	 * Same as {@link #assertNext(String, String, String)} but for a two-columns table.
	 * @param expectedFirst (optional)
	 * @param expectedSecond (optional)
	 * @throws AssertionFailedError
	 * @throws DataSetException
	 */
	public void assertNext(final String expectedFirst, final String expectedSecond) throws AssertionFailedError, DataSetException {
		assertNext(null, new String[]{expectedFirst, expectedSecond});
	}

	/**
	 * Same as {@link #assertNext(String, String, String)} but for a one-column table.
	 * @param expectedFirst (optional)
	 * @throws AssertionFailedError
	 * @throws DataSetException
	 */
	public void assertNext(final String expectedFirst) throws AssertionFailedError, DataSetException {
		assertNext(null, new String[]{expectedFirst});
	}

	/**
	 * Assert that the number of rows in this.{@link #resultTable} is as expected.
	 */
	public void assertRowCount(final int expected) throws AssertionFailedError {
		Assert.assertEquals("There shall be " + expected + " rows in total. The SQL or test data is likely incorrect."
				, expected, resultTable.getRowCount());
	} /* assertRowCount */

	/** Return the internal result table. */
	public ITable getResultTable() {
		return resultTable;
	} /* getResultTable */

}